--[[
	This mod provides several options on how to gift additional tech to players at the beginning of the game, for free.
	By default, this affects human players as well as AI. This can be changed by setting GIFT_HUMAN and/or GIFT_AI to 0.

    The mod uses either the hardcoded approach or a category-based one, both of which are explained in detail below.
    Set USE_HARDCODED_SETS_OF_TECH to true for the hardcoded approach, set it to false for the category-based approach.
    The hardcoded approach will probably be more balanced and better themed when playing as stock races (but you should
    still not expected a fully balanced game), whereas the category-based approach will most likely be significantly
    more wild and random, possibly resulting in very powerful starting advantages.


    Hardcoded approach:

	This approach gifts a set of tech based on the name of the player's race. These tech sets are predetermined for each
	stock race name (such as Alkari). If a custom race name is used, one of several tech sets will be picked at random.

	You can look up (and customize) the gifted tech for stock race names by checking RACE_NAME_TO_GIFTED_STARTING_TECH.
	You can add your own custom race there as well, just keep in mind that each race name is CaSE sEnSitivE.
    To look up and customize the random sets, check HARDCODED_TECH_SETS_TO_RANDOMLY_CHOOSE_FROM_AS_FALLBACK.
    For the sake of fairness, all players (and not just the host) should know the configurations in advance.

	If a gifted tech was already available to the player (due to starting conditions in a late era) then the player will
	be compensated with twice of the RP cost in BC to make up for the 'loss' of this tech.
	This compensation factor can be customized by changing COMPENSATION_FACTOR, but only natural numbers are supported.
	To disable the compensation completely, set COMPENSATE_ALREADY_OWNED_HARDCODED_TECH to false.


    Category-based approach:

    For this approach, each tech is put into one of several categories. Each player is given random techs from these
    categories (by default 1 from each category except CATEGORY_THRIVE). The amount of tech given for each category is
    determined by constants below and can easily be changed,
    e.g. NR_CATEGORY_THRIVE_GIFTED = 2 means each player will randomly draw two tech from CATEGORY_THRIVE.
    The categories are:
    CATEGORY_THRIVE: tech for production, food, research and population as well as expansion, e.g. better fuel cells
    CATEGORY_MISC: tech for money, indirect combat support such as space academy, planetary defense,
                   ground combat including boni for boarding, spying and a few more
    CATEGORY_SHIP_OFFENSE: missiles, torpedos, beams, special weapons including fighters, support systems to improve
                           offense such as computers or fast missile racks
    CATEGORY_SHIP_DEFENSE: all armor tech and shields as well as any other tech that directly helps to keep ships alive
    CATEGORY_SHIP_OTHER: drives, bombs, tech to increase space on ships as well as any other utility that does not
                         mainly fit into offense or defense, such as stasis fields
    You can of course also customize these categories below. You will find that they are constructed by merging several
    sub-categories such as SPY_BOOST which in turn list the individual techs.
    Be mindful when renaming category variable names or when adding new ones, there are several places in the code that
    you will have to adjust accordingly.

    If a player already owns a randomly drawn technology then a redraw will be attempted from the same category.
    If 10000 redraws fail in a row it means the player already owns every technology in the category. This is considered
    an error due to poor configuration and will be reported to the host (there will be no BC compensation).


    General remarks:

    You can find a 'List of all available tech names:' below, avoid typos by copying the values from there.

    You can also disable tech instead of gifting them by adding a prefix (default: !) though this is not currently used
    in any of the default tech sets shipped with this mod.
    E.g. adding "!colony_ship" as a tech will not only take away the ability to build colony ships from the player but
    also prevent him/her from researching them in case of pre-warp. Note that the player might still acquire the tech by
    other means.

    It is highly recommended to check Info->Tech Review at the beginning of the game; some gifted technologies such as
    survival_pods might not be apparent otherwise. No message box listing these techs exists, because clients cannot
    display them and only showing it to the host would give him/her an unfair knowledge-advantage.


	Compatibility:

    Changes to RP cost should be applied before this mod because of the compensation option.

    If you combine it with a mod that removes a certain tech from the tree, you should also remove that tech in this mod
    because it will overwrite disabling it. E.g. by default, phase_shifter can be gifted even when playing 1.50i.


    Changelog:
    - v1: first release, 31th August 2021

	Last update of this mod: 31th August 2021; Tested with version 1.50.18 improved

	Questions etc. to: romanhable@web.de (Epirasque)
]]


GIFT_HUMANS = true
GIFT_AI = true

-- Set to true to give specific sets of tech to players depending on race name (or randomly pick a set from a hard-coded list
-- of tech bundles).
-- Set to false to gift each player technologies from broader categories, drawing each tech individually instead of as a set.
-- The results can be a lot less balanced but obviously allows for more combinations. The used categories can be adjusted.
USE_HARDCODED_SETS_OF_TECH = true

set_seed()

--- SETTINGS FOR GIFTING HARDCODED SETS OF TECH ---
-- Adding this prefix to a tech name will ban it from the player's tech tree instead of gifting it to him/her
COMPENSATE_ALREADY_OWNED_HARDCODED_TECH = true
COMPENSATION_FACTOR = 2 --e.g. factor 2 means the player will get 2*RP where RP refers to tech that was already owned
DISABLE_TECH_PREFIX = "!" --adding this prefix in front of a tech name will cause it to be disabled instead of gifted

-- Note: race names are case-sensitive and should not their length should not exceed 10 characters.
-- To temporarily exclude a race, best comment it out (Race = {} results in 0 techs being gifted to it!)
RACE_NAME_TO_GIFTED_STARTING_TECH = {
    Alkari = {"anti_grav_harness", "sub_space_communications", "spaceport", "inertial_nullifier", "research_laboratory"},
    Bulrathi = {"space_academy", "pleasure_dome", "neural_scanner", "positronic_computer", "advanced_damage_control"},
    Darlok = {"subterranean_farms", "ion_pulse_cannon", "planetary_gravity_generator", "security_stations", "quantum_detonator", "death_spores"},
    Elerian = {"gaia_transformation", "lightning_field", "ion_drive", "space_academy"},
    Gnolam = {"titan_construction", "survival_pods", "telepathic_training", "android_workers", "armor_barracks"},
    Human = {"scout_lab", "quantum_detonator", "iridium_fuel_cells", "transporters", "holo_simulator"},
    Klackon = {"cloning_center", "sub_space_communications", "megafluxers", "fighter_garrison"},
    Meklar = {"recyclotron", "cyber_security_link", "tractor_beam", "anti_matter_bomb"},
    Mrrshan = {"ion_drive", "neutron_scanner", "stealth_suit", "universal_antidote"},
    Psilon = {"terraforming", "telepathic_training", "warp_interdictor"},
    Sakkra = {"battlestation", "neutron_blaster", "biomorphic_fungi", "alien_management_center", "survival_pods", "fusion_rifle"},
    Silicoid = {"tritanium_armor", "planet_construction", "battleoids", "deuterium_fuel_cells"},
    Trilarian = {"pollution_processor", "energy_absorber", "planetary_missile_base", "microbiotics", "warp_dissipator"},
    MyExampleRaceNameWhichIsTooLongToFit = {"terraforming", "planetary_radiation_shield", "ground_batteries", "class_i_shield"}
}
-- One of these sets will be given in case the race name was not specifically listed above
HARDCODED_TECH_SETS_TO_RANDOMLY_CHOOSE_FROM_AS_FALLBACK = {
    {"terraforming", "planetary_radiation_shield", "ground_batteries", "class_i_shield"},
    {"planetary_supercomputer", "optronic_computer", "planetary_stock_exchange"},
    {"automated_repair_unit", "nano_disassemblers", "death_spores", "energy_absorber", "deuterium_fuel_cells"},
    {"thorium_fuel_cells", "fusion_drive", "spaceport", "tachyon_scanner"},
    {"star_gate", "sub_space_teleporter", "planetary_gravity_generator"},
    {"megafluxers", "battle_pods", "subterranean_farms", "biospheres"},
    {"advanced_city_planning", "biospheres", "troop_pods", "fighter_bays"},
    {"multi_phased_shields", "class_i_shield", "planetary_radiation_shield", "personal_shield"},
    {"evolutionary_mutation", "cloning_center", "bio_terminator", "dauntless_guidance_system"},
    {"anti_matter_torpedo", "displacement_device", "pollution_processor"},
    {"doom_star_construction", "xeno_psychology", "microlite_construction"},
    {"planetary_flux_shield", "gyro_destabilizer", "soil_enrichment"},
    {"robo_miners", "automated_repair_unit", "security_stations"}
}

--[[
 List of all available tech names:
  achilles_targeting_unit adamantium_armor advanced_city_planning
  advanced_damage_control alien_management_center android_farmers
  android_scientists android_workers anti_grav_harness anti_matter_bomb
  anti_matter_drive anti_matter_torpedo anti_missile_rocket armor_barracks
  artemis_system_net planet_construction assault_shuttles astro_university
  atmospheric_renewer augmented_engines autolab automated_factories
  automated_repair_unit battleoids battle_pods battle_scanner battlestation
  bio_terminator biomorphic_fungi black_hole_generator bomber_bays capitol
  class_i_shield class_iii_shield class_v_shield class_vii_shield
  class_x_shield cloaking_device cloning_center colony_base colony_ship
  confederation cyber_security_link cybertronic_computer damper_field
  dauntless_guidance_system death_ray death_spores deep_core_mining
  core_waste_dumps deuterium_fuel_cells dimensional_portal displacement_device
  disrupter_cannon doom_star_construction reinforced_hull ecm_jammer
  electronic_computer emissions_guidance_system energy_absorber biospheres
  evolutionary_mutation extended_fuel_tanks fast_missile_racks federation
  fighter_bays fighter_garrison food_replicators freighters fusion_beam
  fusion_bomb fusion_drive fusion_rifle gaia_transformation
  galactic_currency_exchange galactic_cybernet galactic_unification
  gauss_cannon graviton_beam gyro_destabilizer hard_shields heavy_armor
  heavy_fighter_bays heightened_intelligence high_energy_focus holo_simulator
  hydroponic_farm hyper_drive megafluxers hyper_x_capacitors
  hyperspace_communications imperium inertial_nullifier inertial_stabilizer
  interphased_drive ion_drive ion_pulse_cannon iridium_fuel_cells jump_gate
  laser_cannon laser_rifle lightning_field marine_barracks mass_driver
  mauler_device merculite_missile microbiotics microlite_construction
  outpost_ship moleculartronic_computer multi_wave_ecm_jammer
  multi_phased_shields nano_disassemblers neural_scanner neutron_blaster
  neutron_scanner neutronium_armor neutronium_bomb nuclear_bomb nuclear_drive
  nuclear_missile optronic_computer particle_beam personal_shield phase_shifter
  phasing_cloak phasor phasor_rifle planetary_barrier_shield
  planetary_flux_shield planetary_gravity_generator planetary_missile_base
  ground_batteries planetary_radiation_shield planetary_stock_exchange
  planetary_supercomputer plasma_cannon plasma_rifle plasma_torpedo plasma_web
  pleasure_dome pollution_processor positronic_computer powered_armor
  pulse_rifle proton_torpedo psionics pulsar pulson_missile quantum_detonator
  rangemaster_unit recyclotron reflection_field robotic_factory
  research_laboratory robo_miners space_scanner scout_lab security_stations
  sensors shield_capacitor soil_enrichment space_academy spaceport
  spatial_compressor spy_network standard_fuel_cells star_base star_fortress
  star_gate stasis_field stealth_field stealth_suit stellar_converter
  structural_analyzer sub_space_communications sub_space_teleporter
  subterranean_farms survival_pods tachyon_communications tachyon_scanner
  telepathic_training terraforming thorium_fuel_cells time_warp_facilitator
  titan_construction titanium_armor tractor_beam transport transporters
  tritanium_armor troop_pods universal_antidote uridium_fuel_cells
  virtual_reality_network warp_dissipator warp_interdictor
  weather_control_system wide_area_jammer xeno_psychology xentronium_armor
  zeon_missile zortrium_armor adv_biology adv_power adv_physics
  adv_construction adv_fields adv_chemistry adv_computers adv_sociology
]]

-- these are hard-coded in the game and should not be changed
TECH_STATUS_OWNED = 3
TECH_STATUS_RESEARCHABLE = 1
TECH_STATUS_DISABLED = 0

--- SETTINGS FOR GIFTING CATEGORY-BASED TECH ---
-- Define how many random tech of each category will be gifted for a race that is not listed hard-coded above
NR_CATEGORY_THRIVE_GIFTED = 2
NR_CATEGORY_MISC_GIFTED = 1
NR_CATEGORY_SHIP_OFFENSE_GIFTED = 1
NR_CATEGORY_SHIP_DEFENSE_GIFTED = 1
NR_CATEGORY_SHIP_OTHER_GIFTED = 1

function merge_tables(table1, table2) -- arbitrary nr. of parameters using (...) with arg apparently not supported
    local result = {table.unpack(table1)}
    table.move(table2, 1, #table2, #result+1, result)
    return result
end

-- Specifically excluded from random categories: pre-warp tech (nuclear_missile, colony_ship etc.), antaran tech
-- (except quantum_detonator), government-upgrades such as federation (might not work properly), dimensional_portal
-- and any hyper-advanced research
COLONY_BOOST = {"advanced_city_planning", "android_workers", "android_farmers", "astro_university", "atmospheric_renewer",
                "automated_factories", "biomorphic_fungi", "cloning_center", "deep_core_mining", "core_waste_dumps", "biospheres", "food_replicators",
                "gaia_transformation", "holo_simulator", "hydroponic_farm", "microbiotics", "microlite_construction", "nano_disassemblers",
                "planetary_gravity_generator", "pleasure_dome", "pollution_processor", "recyclotron", "robotic_factory", "robo_miners",
                "soil_enrichment", "subterranean_farms", "terraforming", "universal_antidote", "virtual_reality_network", "weather_control_system"}
RESEARCH_BOOST = {"autolab", "heightened_intelligence", "galactic_cybernet", "android_scientists", "planetary_supercomputer", "research_laboratory", "scout_lab"}
MONEY_BOOST = {"galactic_currency_exchange", "planetary_stock_exchange", "spaceport"}
EXPANSION = {"planet_construction", "deuterium_fuel_cells", "iridium_fuel_cells", "thorium_fuel_cells", "uridium_fuel_cells"}
SPY_BOOST = {"cyber_security_link", "neural_scanner", "psionics", "stealth_suit", "telepathic_training"}
INDIRECT_COMBAT_SUPPORT = {"advanced_damage_control", "alien_management_center", "hyperspace_communications", "jump_gate", "space_academy", "star_gate",
                           "stealth_field", "sub_space_communications", "survival_pods", "tachyon_communications", "warp_interdictor"}
GROUND_COMBAT_BOOST = {"anti_grav_harness", "fusion_rifle", "personal_shield", "phasor_rifle", "plasma_rifle", "powered_armor"}
GROUND_COMBAT_DEFENSE = {"armor_barracks", "battleoids"}
PLANETARY_DEFENSE = {"artemis_system_net", "battlestation", "fighter_garrison", "planetary_barrier_shield", "planetary_flux_shield", "planetary_missile_base",
                     "ground_batteries", "planetary_radiation_shield", "star_fortress"} --planetary_radiation_shield kept here instead of colony_boost for consistency with other planetary shields
BOMBS = {"anti_matter_bomb", "bio_terminator", "death_spores", "fusion_bomb", "neutronium_bomb"}
SHIP_COMBAT_MOVEMENT = {"anti_matter_drive", "fusion_drive", "hyper_drive", "augmented_engines", "interphased_drive", "ion_drive", "sub_space_teleporter",
                        "inertial_nullifier", "inertial_stabilizer"}
SHIP_MISSILES_TORPEDOS = {"anti_matter_torpedo", "merculite_missile", "plasma_torpedo", "proton_torpedo", "pulson_missile", "zeon_missile"}
SHIP_BEAMS = {"disrupter_cannon", "fusion_beam", "gauss_cannon", "graviton_beam", "ion_pulse_cannon", "mass_driver", "mauler_device", "neutron_blaster", "phasor",
              "plasma_cannon"}
SHIP_SPECIAL_WEAPONS = {"gyro_destabilizer", "plasma_web", "pulsar", "stellar_converter"} --stellar_converter fits better here than in planetary defense
SHIP_OFFENSE_SUPPORT = {"achilles_targeting_unit", "dauntless_guidance_system", "emissions_guidance_system", "fast_missile_racks", "high_energy_focus", "hyper_x_capacitors",
                        "battle_scanner", "cybertronic_computer", "moleculartronic_computer", "optronic_computer", "positronic_computer", "rangemaster_unit",
                        "neutron_scanner", "sensors", "structural_analyzer", "tachyon_scanner"}
CARRIED_SHIPS = {"assault_shuttles", "bomber_bays", "fighter_bays", "heavy_fighter_bays"}
ARMOR = {"adamantium_armor", "neutronium_armor", "tritanium_armor", "zortrium_armor"}
SHIP_DEFENSE = {"anti_missile_rocket", "automated_repair_unit", "class_i_shield", "class_iii_shield", "class_vii_shield", "class_x_shield", "displacement_device",
                "reinforced_hull", "ecm_jammer", "energy_absorber", "hard_shields", "heavy_armor", "lightning_field", "multi_wave_ecm_jammer", "multi_phased_shields",
                "security_stations", "shield_capacitor", "cloaking_device", "phasing_cloak", "wide_area_jammer"}
SHIP_UTILITY = {"phase_shifter", "stasis_field", "time_warp_facilitator", "tractor_beam", "troop_pods", "transporters", "warp_dissipator", "quantum_detonator"}
SHIP_SPACE_BOOST = {"battle_pods", "doom_star_construction", "megafluxers", "titan_construction"}
OTHER = {"xeno_psychology", "evolutionary_mutation"}


CATEGORY_THRIVE = merge_tables(merge_tables(
        COLONY_BOOST, EXPANSION), RESEARCH_BOOST)
CATEGORY_MISC = merge_tables(merge_tables(merge_tables(merge_tables(merge_tables(merge_tables(
        MONEY_BOOST, INDIRECT_COMBAT_SUPPORT), GROUND_COMBAT_BOOST), GROUND_COMBAT_DEFENSE), PLANETARY_DEFENSE), SPY_BOOST), OTHER)
CATEGORY_SHIP_OFFENSE = merge_tables(merge_tables(merge_tables(merge_tables(
        SHIP_MISSILES_TORPEDOS, SHIP_BEAMS), SHIP_SPECIAL_WEAPONS), SHIP_OFFENSE_SUPPORT), CARRIED_SHIPS)
CATEGORY_SHIP_DEFENSE = merge_tables(
        ARMOR, SHIP_DEFENSE)
CATEGORY_SHIP_OTHER = merge_tables(merge_tables(merge_tables(
        SHIP_COMBAT_MOVEMENT, BOMBS), SHIP_UTILITY), SHIP_SPACE_BOOST)

--- MOD LOGIC ---
function gift_tech_of_category(category, amount_to_gift, player_techs)
    local nr_gifted = 0
    local nr_tries = 0
    while (nr_gifted < amount_to_gift and nr_tries < 10000) do
        nr_tries = nr_tries + 1
        local candidate_to_gift = category[random(#category)]
        local player_tech_status = player_techs[candidate_to_gift]
        if player_tech_status == TECH_STATUS_RESEARCHABLE then
            player_techs[candidate_to_gift] = TECH_STATUS_OWNED
            nr_gifted = nr_gifted + 1
        end
    end
    if nr_tries >= 10000 then
        msgbox("TECH_STA: ERROR: Stopped granting techs after 10000 tries -> player probably already has all of the given category:\n" .. tostring(category))
    end
    return player_techs
end

if(USE_HARDCODED_SETS_OF_TECH) then
    output =           "TECH_STA: Gifting players additional predetermined starting tech depending on name of picked race.\n"
    output = output .. "Race names not on the list will be given a random set of tech out of " .. tostring(#HARDCODED_TECH_SETS_TO_RANDOMLY_CHOOSE_FROM_AS_FALLBACK) .. " predetermined sets.\n"
    output = output .. "Each tech that was already available due to era will be compensated with " .. tostring(COMPENSATION_FACTOR) .. "*RP in BC."
    msgbox(output)
else
    output =           "TECH_STA: Gifting players additional starting tech, randomly picking them individually for every player from categories below:\n"
    output = output .. "\nEMPIRE:\n"
    output = output .. NR_CATEGORY_THRIVE_GIFTED .. " tech for thriving\n"
    output = output .. NR_CATEGORY_MISC_GIFTED .. " other tech\n"
    output = output .. "\nSHIP COMBAT:\n"
    output = output .. NR_CATEGORY_SHIP_OFFENSE_GIFTED .. " tech for ship offense\n"
    output = output .. NR_CATEGORY_SHIP_DEFENSE_GIFTED .. " tech for ship defense\n"
    output = output .. NR_CATEGORY_SHIP_OTHER_GIFTED .. " other tech\n"
    msgbox(output)
end

nr_custom_race_names = 0
for player_id, player in pairs(get_players()) do
    collectgarbage()
    if((player.is_human == 1 and GIFT_HUMANS == true) or (player.is_human == 0 and GIFT_AI == true)) then
        local player_techs = get_player_techs(player_id)
        local player_compenstation = 0
        if(USE_HARDCODED_SETS_OF_TECH == true) then
            local tech_set_to_give = RACE_NAME_TO_GIFTED_STARTING_TECH[player.race]
            if tech_set_to_give == nil then
                nr_custom_race_names = nr_custom_race_names + 1
                tech_set_to_give = HARDCODED_TECH_SETS_TO_RANDOMLY_CHOOSE_FROM_AS_FALLBACK[random(#HARDCODED_TECH_SETS_TO_RANDOMLY_CHOOSE_FROM_AS_FALLBACK)]
            end
            for tech_id, tech in pairs(tech_set_to_give) do
                if tech:sub(0, #DISABLE_TECH_PREFIX) == DISABLE_TECH_PREFIX then
                    local tech_to_disable = tech:sub(#DISABLE_TECH_PREFIX + 1, #tech)
                    player_techs[tech_to_disable] = TECH_STATUS_DISABLED
                else
                    local current_tech_state = player_techs[tech]
                    if current_tech_state == nil then
                        msgbox("TECH_STA: ERROR: unknown tech will be ignored: '" .. tostring(tech) .. "' maybe a typo?")
                    else
                        if current_tech_state == TECH_STATUS_OWNED and COMPENSATE_ALREADY_OWNED_HARDCODED_TECH == true then
                            local techs = rget_techs()
                            local fields = rget_fields()
                            local rp_cost = fields[techs[tech].field].cost
                            player_compenstation = player_compenstation + COMPENSATION_FACTOR*rp_cost
                        else
                            player_techs[tech] = TECH_STATUS_OWNED
                        end
                    end
                end
            end
        else
            gift_tech_of_category(CATEGORY_THRIVE, NR_CATEGORY_THRIVE_GIFTED, player_techs)
            gift_tech_of_category(CATEGORY_MISC, NR_CATEGORY_MISC_GIFTED, player_techs)
            gift_tech_of_category(CATEGORY_SHIP_OFFENSE, NR_CATEGORY_SHIP_OFFENSE_GIFTED, player_techs)
            gift_tech_of_category(CATEGORY_SHIP_DEFENSE, NR_CATEGORY_SHIP_DEFENSE_GIFTED, player_techs)
            gift_tech_of_category(CATEGORY_SHIP_OTHER, NR_CATEGORY_SHIP_OTHER_GIFTED, player_techs)
        end
        local rplayers = rget_players()
        rplayers[player_id].bc = rplayers[player_id].bc + player_compenstation
        rset_players(rplayers)
        set_player_techs(player_id, player_techs)
    end
end

if(USE_HARDCODED_SETS_OF_TECH == true) then
    msgbox("TECH_STA: Number of players who start with a randomly determined tech set due to their race name not being stock: " .. tostring(nr_custom_race_names))
end