-- Range that does not cause any tech to cap at 65536 RP: 0 to 436 
FIRST_BOUND_PERCENT  = 80

-- Range that does not cause any tech to cap at 65536 RP: 0 to 436 
SECOND_BOUND_PERCENT = 120

--[[
	This mod varies the cost of each research field separately. 
	By default, the new cost will be randomly up to 20% bigger or smaller compared to the 
	original cost. 
	
	You can easily configure this by changing the values for FIRST_BOUND_PERCENT and 
	SECOND_BOUND_PERCENT which define the interval for the variation. 
	
	Some examples:
	
	Default:
	FIRST_BOUND_PERCENT  =  80
	SECOND_BOUND_PERCENT = 120
	-> each research field randomly has 80% to 120% of its default cost
	
	Exactly half cost:
	FIRST_BOUND_PERCENT  = 50
	SECOND_BOUND_PERCENT = 50

	Exactly double cost:
	FIRST_BOUND_PERCENT  = 200
	SECOND_BOUND_PERCENT = 200

	Most extreme variation:
	FIRST_BOUND_PERCENT = 0
	SECOND_BOUND_PERCENT = 436
	-> each research field randomly has 0% to 436% of the default cost
	
	Please Note:
	
	Only integer values are supported at the moment. 
	
	The costs are capped at 65536 RP to avoid an issue with ship design screen. 
	A warning will pop up if this occurs, showing you how many research fields 
	are affected. Hyper advanced fields are not capped but can produce negative 
	cost-values if the bounds are too extreme. 
	
	Hyper advanced tech are varied as well, but only once for all of these tech. 
	Researching the next field beyond the first hyper advanced tech of a category 
	will increase the cost by a static amount as set in the config (default: 10000 RP more)
	
	It does not matter if FIRST_BOUND_PERCENT or SECOND_BOUND_PERCENT has the bigger value. 
	
	Known issue: In singleplayer, the cost values are not reset when you 
	start a new game, meaning the already varied cost will be varied even 
	more. To avoid this, restart the game before creating a new one. 
	
	For this mod, significant amounts of code were reused from the RANDTECH mod of 1.50.18.
	
	Last update of this mod: 20th February 2021; Tested with version 1.50.13 improved
	
	Questions etc. to: romanhable@web.de (Epirasque)
]]

if FIRST_BOUND_PERCENT < 0 or SECOND_BOUND_PERCENT < 0 or type(FIRST_BOUND_PERCENT) ~= type(0) or type(SECOND_BOUND_PERCENT) ~= type(0) then
	msgbox("Invalid configuration for TECH_VAR: bounds have to be non-negative numbers!")
	msgbox(type(FIRST_BOUND_PERCENT))
	goto skip_tech_var
end

if FIRST_BOUND_PERCENT > SECOND_BOUND_PERCENT then
	local tmp = FIRST_BOUND_PERCENT
	FIRST_BOUND_PERCENT = SECOND_BOUND_PERCENT
	SECOND_BOUND_PERCENT = tmp
end

advf = {
    advf_biology = 1,
    advf_power = 1,
    advf_physics = 1,
    advf_construction = 1,
    advf_fields = 1,
    advf_chemistry = 1,
    advf_computers = 1,
    advf_sociology = 1
}

function set_id(t, tag)
    for i,v in pairs(t) do
        v[tag] = i
    end
end

function make_branches(ff, keep_gray_boxes)
    local branches = {}
    local f2br = {}
    local starting_fields = rget_starting_fields()
	local nr_capped = 0

    local map = { engineering = "advanced_engineering" }
    if keep_gray_boxes then
        map.chemistry = "advanced_metallurgy"
        map.nuclear_fission = "advanced_fusion"
        map.physics = "fusion_physics"
        map.electronics = "optronics"
    end
    for i,s in pairs(starting_fields) do
        if map[s] ~= nil then
            starting_fields[i] = map[s]
        end
    end

    for i,f in pairs(starting_fields) do
        local n = rawlen(branches) + 1
        local b = {
            branch_last_field = "advf_" .. i,
            branchnum = n,
            field = f,
        }
        table.insert(branches, b)
        ff[f].is_branch_root = 1
    end

    for i,b in ipairs(branches) do
        local fname = b.field
        local marks = {}
        local n = 1
        while advf[fname] == nil and ff[fname] ~= nil do
            local f = ff[fname]
            if marks[fname] ~= nil then
                error(tostring("detected tech loop ", marks))
            end
            marks[fname] = n
            n = n + 1
            f.branch = i
            --f.techs = {}
			fname = f.next_field
        end
    end

    return branches
end

function generate_variation_factor_5_digits()
	-- neither math library nor float values are supported yet in raw mode -> use big integers to avoid too much rounding; random starts from 1 not 0
	return random(100 * (SECOND_BOUND_PERCENT - FIRST_BOUND_PERCENT) + 1 ) - 1 + 100 * FIRST_BOUND_PERCENT
end

function varyFieldCost(fieldToVary)
	if fieldToVary.cost == nil then
		return 0
	end
	local had_to_cap = 0

	local variation_factor_5_digits = generate_variation_factor_5_digits()
	
	fieldToVary.cost = (fieldToVary.cost * variation_factor_5_digits) / 10000
	if fieldToVary.cost > 65536 then
		if advf[fieldToVary.name] == nil then
			fieldToVary.cost = 65536
			had_to_cap = 1
		end
	end
	
	return had_to_cap
end

function go()
    set_seed()
    local ff = rget_fields() set_id(ff, "name")
    local tt = rget_techs()  set_id(tt, "name")
	local bb = make_branches(ff, 1)
	
	local nr_capped = 0
	for _, field in pairs(ff) do
		nr_capped = nr_capped + varyFieldCost(field)
	end
	
    rset_fields(ff)
    rset_techs(tt)
	
	output = "TECH_VAR:\n\n"
	
	if FIRST_BOUND_PERCENT ~= SECOND_BOUND_PERCENT then
		output = output .. "Research field costs have been varied to end up randomly inbetween " .. FIRST_BOUND_PERCENT .. "% and " .. SECOND_BOUND_PERCENT .. "% of the original RP cost, separately rolled for every research field."
	else
		output = output .. "Research field costs have been set to " .. FIRST_BOUND_PERCENT .. "% of their original RP cost value."
	end

	if nr_capped > 0 then
		output = output .. "\n\nWARNING: Capped " .. nr_capped .. " fields to 65536 RP to avoid ship design screen issues."
	end
	
	msgbox(output)

end

go()
::skip_tech_var::
